' Minimal SD booter Test
' ----------------------
' (c)2012 Andy Schenk, Insonix
' RR20120820 modified by Cluso99 to test SD cards (for later use in P2 ROM)
' RR20120824 display a few more details on MBR


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

   do  = 0      'sd pins
   clk = 1
   di  = 2
   cs  = 3


'  do  = 8      'sd pins
'  clk = 9
'  di  = 10
'  cs  = 11

{{
'RamBlade3
  _clkmode  = xtal1 + pll16x
  _xinfreq  = 6_500_000
  DO     = 20                'microSD
  Clk    = 21
  DI     = 22
  CS     = 23
}}

obj  dbg: "FullDuplexSerial"

var
   byte   done[4]        'pasm sets done[1] to $0 when done
   byte   buff[512]      'pasm stores the MBR sector here
   byte   param[512]     'pasm stores some info here
   byte   overflo[16384] 'allow big buffer in case of slower card
   
pub Main : i | tm
'
' This Spin code is not part of the boot code, it
' just starts the PASM cog and shows the result
'
  waitcnt(clkfreq*3+cnt)
  dbg.start(31,30,0,115200)
  bytefill(@done,0,4+512+512)  'clear the buffers
  done[0] := $FF                'set busy

  cognew(@entry, @done)

  dbg.str(string("testing...",13))
  repeat until (done[0] == 0)

  DisplayParam
  DisplayMBR
  
  repeat


Pri DisplayParam | i, j
  dbg.str(string("Parameters used...",13))
  repeat j from 0 to 31          'just display 32*16 = 512 bytes
    repeat i from 0 to 15
      dbg.hex(param[i + j*16],2)
      dbg.tx(" ")
    dbg.tx(13)

Pri DisplayMBR | r, i, j
  dbg.str(string("MBR sector 0...",13))
  repeat j from 0 to 31
    repeat i from 0 to 15
      dbg.hex(buff[i + j*16],2)
      dbg.tx(" ")
    dbg.tx(" ")
    repeat i from 0 to 15
      if buff[i+ j*16] < $20 or buff[i+ j*16] > $7F
        dbg.tx(".")
      else
        dbg.tx(buff[i + j*16])
    dbg.tx(13)
'display the 16 bytes we might use immediately preceeding the ptn tables
  dbg.str(string("$1AE(16): "))
  repeat i from 0 to 15
    dbg.hex(buff[$1AE + i],2)
  dbg.tx(13)
'display the ptn tables if it is an MBR
  if buff[510] == $55 and buff[511] == $AA
    dbg.str(string("Valid MBR",13))
    repeat j from 0 to 3
      dbg.str(string("PTN "))
      dbg.dec(j)
      dbg.str(string(": Typ=$"))
      dbg.hex(buff[$1BE+$04+ j*16],2)
      dbg.tx(13)
  else
    dbg.str(string("Not MBR",13))

dat
        org
entry
{
        mov   hubparam, par     '@param
        mov   hubbuf,par
        add   hubbuf, sectorsz  'offset to @buff (@param +512)
        mov   hubdone, hubbuf
        add   hubdone, sectorsz 'offset to @done (@param +512 +512)
}
        mov   hubdone, par     '@done
        mov   hubbuf, hubdone
        add   hubbuf, #4        'offset to @buff (@done +4)
        mov   hubparam,hubbuf
        add   hubparam,sectorsz 'offset to @param (@done +4+512)
        
' set up
        mov blocksh,#0
        or outa,clkmask
        mov dira,outmask        'pins
        neg phsb,#1
        mov ctr2,onek           '1000 x High for init
oneloop
        call #sendiohi
        djnz ctr2,#oneloop
        mov starttime,cnt
         mov ctr2,#9           'try a few times
reslp   mov cmdo,#0
        mov cmdp,#0
        call #cmd
         cmp accb,#1  wz
  if_z   jmp #tsthc
         djnz ctr2,#reslp
tsthc    mov cmdo,#8           'test if SD2
         mov cmdp,#$1AA
         call #cmd
         cmp accb,#1  wz
  if_nz  jmp #resok            '< no
         mov ctr,#32
         call #sendiocnt       'read volt
         
inilp    mov cmdo,#55
         mov cmdp,#0
         call #cmd
         mov cmdo,#41
         mov cmdp,bit30
         call #cmd
         cmp accb,#0  wz
   if_nz jmp #inilp
   
         mov cmdo,#58         'test if SDHC
         mov cmdp,#0
         call #cmd
         mov ctr,#32
         call #sendiocnt
         or outa,csmask
         test accb,bit30  wz
   if_z  mov blocksh,#9      
         jmp #begin
resok   or outa,csmask
        call #sendiohi
initloop
        mov cmdo,#55
        call #cmd
        mov cmdo,#41
        call #cmd
        cmp accb,#1 wz
   if_z jmp #initloop
        mov blocksh,#9         '0=SDHC 9=SD

' read MBR sector#0
begin   mov  blocknr,#0
        mov  bufad,hubbuf
        call #rblock           'read sector 0
fail
        wrlong x00, hubdone    'set done       
halt    jmp  #halt

        
rblock
        mov  ctr2,sectorsz     'read block
        mov starttime,cnt
        mov cmdo,#17
        mov  cmdp,blocknr
        call #cmd
        call #readresp
rbyte
        mov  accb,#8
        andn outa,clkmask
        or   outa,clkmask
        test domask,ina  wc
        addx acca,acca
        djnz accb,#rbyte+1
        wrbyte acca,bufad
        add  bufad,#1
        djnz ctr2,#rbyte        
        call #sendiohi
        call #sendiohi
        or outa,csmask
rblock_ret  ret

sendio
        rol phsb,#24            'SPI routines
sendiohi
        mov ctr,#8
sendiocnt
        mov accb,#0
bit     andn outa,clkmask
        rol phsb,#1  wc
        muxc outa,dimask
        or  outa,clkmask
        test domask,ina wc
        addx accb,accb        
        djnz ctr,#bit
sendio_ret
sendiohi_ret
sendiocnt_ret
        ret

checktime
        mov duration,cnt
        sub duration,starttime
        cmp duration,clockfreq wc
checktime_ret
  if_c  ret
        neg accb,#13          'timeout error: -13

        wrbyte accb, hubparam
        add   hubparam, #1
        jmp   #fail        
        
cmd
        wrbyte cmdo, hubparam
        add   hubparam, #1
        wrbyte cmdp, hubparam
        add   hubparam, #1
        
         cmp cmdo,#41  wz
  if_z   or outa,csmask
         nop
        andn outa,csmask
        neg phsb,#1
        call #sendiohi
        mov phsb,cmdo
        add phsb,#$40
        call #sendio
        mov phsb,cmdp
        shl phsb,blocksh
        call #sendiohi
        call #sendiohi
        call #sendiohi
        call #sendiohi
        cmp cmdo,#8  wz
  if_z  mov phsb,#$87
  if_nz mov phsb,#$95
        call #sendio
        cmp cmdo,#12 wz
  if_z  neg phsb,#1
  if_z  call #sendiohi
readresp
        neg phsb,#1
        call #sendiohi
        call #checktime
        cmp accb,#$ff wz
   if_z jmp #readresp

        wrbyte  accb, hubparam
        add     hubparam, #1
        wrbyte  xff, hubparam
        add     hubparam, #1
         
cmd_ret
readresp_ret
        ret


clockfreq  long 80_000_000
onek       long 1000
sectorsz   long 512
bit30      long 1<<30

domask     long 1<<do
csmask     long 1<<cs
clkmask    long 1<<clk
dimask     long 1<<di
outmask    long 1<<clk | 1<<di | 1<<cs

lastblock  long 19000
signatur0    byte "BOOT"
signaturFile byte "P2BO"
'hubbuf     long $7000
hubparam      long      0
hubbuf        long      0
hubdone       long      0
xff           long      $FF
x00           long      0

acca    res 1
accb    res 1
cmdo    res 1
cmdp    res 1
ctr     res 1
ctr2    res 1
starttime res 1
duration res 1

bufad    res 1
blocknr  res 1
blocksh  res 1